/*
 * ******************************************************************************************
 *
 * File		: apb_spi_flash_driver.c
 * Author	: GOWIN
 * Function	: Driver definition of APB SPI Flash
 *
 * ******************************************************************************************
 */

// Includes ---------------------------------------------------------------------------------
#include "apb_spi_flash_driver.h"


// Definitions ------------------------------------------------------------------------------

static unsigned char apb_spi_flash_get_fifo_depth(unsigned char fifo_depth_config);
static void apb_spi_flash_read_status(APB_SPI_RegDef *DEV_FLASH);
static void apb_spi_flash_write_cmd(APB_SPI_RegDef *DEV_FLASH, unsigned int cmd);


// Initializes APB SPI Flash
int apb_spi_flash_init(APB_SPI_RegDef *DEV_FLASH)
{
	unsigned int buff;

	DEV_FLASH->CTRL |= 0x01;         					// Reset APB SPI Flash
	while((DEV_FLASH->CTRL & 0x01)); 					// Wait until APB SPI Flash reset complete

	buff = DEV_FLASH->CONFIG;
	unsigned char tx_fifo_depth_config = (buff & 0x3f) >> 4;
	unsigned char tx_fifo_depth = apb_spi_flash_get_fifo_depth(tx_fifo_depth_config);
	unsigned char rx_fifo_depth_config = (buff & 0x03);
	unsigned char rx_fifo_depth = apb_spi_flash_get_fifo_depth(rx_fifo_depth_config);

	while((DEV_FLASH->STATUS & 0x01)); 					// Wait until APB SPI Flash active finish

	buff = (2 << 16) +(7 << 8) + (0 << 7);
	// AddrLen = 3; DataLen = 8 DataMerge = 0
	// Other fields reset to 0 while spi_default_modes connected to 0
	DEV_FLASH->TRANSFMT = buff;

	DEV_FLASH->CTRL |= (0x1 << 2);    					// Reset TX FIFO
	while((DEV_FLASH->CTRL & 0x2));  					// Wait TX FIFO reset complete
	DEV_FLASH->CTRL  |= ((tx_fifo_depth/2) << 16);  	// Set TX threshold as half of TX FIFO depth

	DEV_FLASH->CTRL |= (0x1 << 1);    					// Reset RX FIFO
	while((DEV_FLASH->CTRL & 0x1));  					// Wait RX FIFO reset complete
	DEV_FLASH->CTRL  |= ((rx_fifo_depth/2) << 8);   	// Set RX threshold as half of RX FIFO depth

	// Do not support Interrupt signal
	DEV_FLASH->INTREN &= 0x00000000;        			// Close all the interrupt
	DEV_FLASH->TIMING &= (~(0xff));
	DEV_FLASH->TIMING |= (0x4);        					// Set the APB SPI Flash interface timing register (0x40)
	                                   	   	   	   		// to set SCLK_DIV to 4, which divide the spi_clk into
	                                   	   	   	   		// ((4+1)*2) = 10 as SCLK

	return APB_SPI_DRIVER_OK;
}

// Switch APB SPI Flash mode between download and read/write/erase memory
int apb_spi_flash_switch_mode(APB_SPI_RegDef *DEV_FLASH)
{
	unsigned int i;

	// Read APB SPI Flash memory access register 0x50
	i = DEV_FLASH->MEMCTRL;

	// Write back to APB SPI Flash register
	DEV_FLASH->MEMCTRL = i;

	// Wait the memCtrlChg become 0
	while(DEV_FLASH->MEMCTRL & (0x01 <<8 ));

	return APB_SPI_DRIVER_OK;
}

// Read APB SPI Flash data
int apb_spi_flash_read(APB_SPI_RegDef *DEV_FLASH, unsigned int len, unsigned int cmd, unsigned int address, unsigned char *buf)
{
	if(buf == NULL)
	{
		return APB_SPI_DRIVER_ERROR;
	}

	DEV_FLASH->TRANSCTRL =   (1 << 30) +        // Enable command
						     (1 << 29) +        // Open the address
			                 (2 << 24) +        // Transfer mode = 2 (read only)
						     (0 << 22) +        // Dual mode 0 regular mode
						     ((len-1) << 0);    // Set read transfer byte count

	DEV_FLASH->ADDR = address;
	DEV_FLASH->CMD  = cmd;

	for(unsigned int i = 0; i < len; i++)
	{
		while(DEV_FLASH->STATUS & (0x1 <<14));	// Wait if RX FIFO is empty

		*buf = DEV_FLASH->DATA;
		buf++;
	}

	while((DEV_FLASH->STATUS & 0x01)); 		 	// Wait until APB SPI Flash active finish

	return APB_SPI_DRIVER_OK;
}

// Write APB SPI Flash data
int apb_spi_flash_write(APB_SPI_RegDef *DEV_FLASH, unsigned int len, unsigned int cmd, unsigned int address, unsigned char *buf)
{
	if(buf == NULL)
	{
		return APB_SPI_DRIVER_ERROR;
	}

	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_ENABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);

	DEV_FLASH->TRANSCTRL =   (1 << 30) +        	// Enable command
						     (1 << 29) + 			// Enable address
						     (1 << 24) +        	// Transfer mode = 1 (write only)
						     (0 << 22) +         	// Dual mode
						     ((len-1) << 12); 		// Set read transfer byte count

	DEV_FLASH->ADDR = address;
	DEV_FLASH->CMD  = cmd;

	for(unsigned int i = 0; i < len; i++)
	{
		while(DEV_FLASH->STATUS & (0x1 <<23));		// Check the status of TX FIFO

		DEV_FLASH->DATA = (unsigned int)(*buf);
		buf++;
	}

	while((DEV_FLASH->STATUS & 0x01)); 				// Wait until APB SPI Flash active finish
	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_DISABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);

	return APB_SPI_DRIVER_OK;
}

// Program APB SPI Flash data
int apb_spi_flash_program(APB_SPI_RegDef *DEV_FLASH, unsigned int len, unsigned int address, unsigned char *buf)
{
	if(buf == NULL)
	{
		return APB_SPI_DRIVER_ERROR;
	}

	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_ENABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);

	DEV_FLASH->TRANSCTRL =   (1 << 30) +           // Enable command
						     (1 << 29) +           // Enable address
			                 (1 << 24) +           // Transfer mode = 1 (write only)
						     (0 << 22) +           // Dual mode
						     ((len-1) << 12);      // Set read transfer byte count

	DEV_FLASH->ADDR = address;
	DEV_FLASH->CMD  = APB_SPI_FLASH_PROGRAM_CMD; 	// Page program command

	for(unsigned int i = 0; i < len; i++)
	{
		while(DEV_FLASH->STATUS & (0x1 <<23));		// Check the status of TX FIFO

		DEV_FLASH->DATA = (unsigned int)(*buf);
		buf++;
	}

	while((DEV_FLASH->STATUS & 0x01)); 		   	   // Wait until APB SPI Flash active finish

	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_DISABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);    		// Wait APB SPI Flash free

	return APB_SPI_DRIVER_OK;
}

// Erase 4K sector
int apb_spi_flash_erase_4ksec(APB_SPI_RegDef *DEV_FLASH, unsigned int address)
{
	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_ENABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);

	// Little endian and the range of address
	unsigned char temp[3];
	DEV_FLASH->TRANSCTRL =   (1 << 30) +        	// Enable command
			                 (1 << 24) +        	// Transfer mode = 1 (write only)
						     (0 << 22) +        	// Dual mode
						     (2 << 12);         	// Set address - 3bytes

	temp[0] = (address >> 16 &0xFF);		    	// The address High byte
	temp[1] = (address >> 8  &0xFF);		    	// The address middle byte
	temp[2] = (address       &0xFF);		    	// The address low   byte

	DEV_FLASH->CMD  = APB_SPI_FLASH_ERASE_4K_CMD;   //4K sector erase command

	for(unsigned char i = 0; i < 3; i++)
	{
		while(DEV_FLASH->STATUS & (0x1 <<23));		// Check the status of TX FIFO

		DEV_FLASH->DATA = temp[i];
	}

	while((DEV_FLASH->STATUS & 0x01)); 				// Wait until APB SPI Flash active finish

	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_DISABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);    		// Wait APB SPI Flash free

	return APB_SPI_DRIVER_OK;
}

// Erase 64K sector
int apb_spi_flash_erase_64ksec(APB_SPI_RegDef *DEV_FLASH, unsigned int address)
{
	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_ENABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);

	// Little endian and the range of address
	unsigned char temp[3];
	DEV_FLASH->TRANSCTRL =   (1 << 30) +        		// Enable command
			                 (1 << 24) +        		// Transfer mode = 1 (write only)
						     (0 << 22) +        		// Dual mode
						     (2 << 12);         		// Set address - 3bytes

	temp[0] = (address >> 16 &0xFF);		    		// The address High byte
	temp[1] = (address >> 8  &0xFF);		    		// The address middle byte
	temp[2] = (address       &0xFF);		    		// The address low byte

	DEV_FLASH->CMD  = APB_SPI_FLASH_ERASE_64K_CMD;		// 64K sector erase command

	for(unsigned char i = 0; i < 3; i++)
	{
		while(DEV_FLASH->STATUS & (0x1 <<23));			// Check the status of TX FIFO

		DEV_FLASH->DATA = temp[i];
	}

	while((DEV_FLASH->STATUS & 0x01)); 					// Wait until APB SPI Flash active finish

	apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_DISABLE_CMD);
	apb_spi_flash_read_status(DEV_FLASH);    			// Wait APB SPI Flash free

	return APB_SPI_DRIVER_OK;
}

// Get APB SPI Flash FIFO depth
static unsigned char apb_spi_flash_get_fifo_depth(unsigned char fifo_depth_config)
{
	unsigned char fifo_depth = 1;

	for(int i = 0; i < fifo_depth_config+1; i++)
	{
		fifo_depth *= 2;
	}

	return fifo_depth;
}

// Write APB SPI Flash command
static void apb_spi_flash_write_cmd(APB_SPI_RegDef *DEV_FLASH, unsigned int cmd)
{
	DEV_FLASH->TRANSCTRL =   (0 << 30) +        		// Disable command
			                 (1 << 24) +        		// Transfer mode = 1 (write only)
						     (0 << 22) +        		// Regular mode
						     (0 << 12);         		// Set transfer byte count is 1
	DEV_FLASH->CMD  = APB_SPI_FLASH_WRITE_CMD;			// Start to transfer

	if(cmd == APB_SPI_FLASH_ERASE_CHIP_CMD)
	{
		apb_spi_flash_write_cmd(DEV_FLASH, APB_SPI_FLASH_WRITE_ENABLE_CMD);
		apb_spi_flash_read_status(DEV_FLASH);
	}

	// Check the status of TX FIFO
	while(DEV_FLASH->STATUS & (0x1 <<23));

	DEV_FLASH->DATA = cmd;

	// Check the status of Transfer
	while((DEV_FLASH->STATUS & 0x01));
}

// Read APB SPI Flash status
static void apb_spi_flash_read_status(APB_SPI_RegDef *DEV_FLASH)
{
	unsigned char temp_first;

	DEV_FLASH->TRANSCTRL =   (1 << 30) +        	// Enable command
			                 (2 << 24) +        	// Transfer mode = 2 (read only)
						     (0 << 22) +        	// Dual mode 0 regular mode
						     (0 << 0);          	// Set read 1 byte

   do
   {
	   	DEV_FLASH->CMD  = APB_SPI_FLASH_READ_STATUS_CMD;	// Read status
		temp_first  = DEV_FLASH->DATA;
   } while(temp_first & 0x01 );				    	// Select the low bit and wait become 0

   while((DEV_FLASH->STATUS & 0x03)); 	    		// Wait until APB SPI Flash active finish
}
