/*
 * ******************************************************************************************
 *
 * File		: apb_spi_driver.c
 * Author	: GOWIN
 * Function	: Driver definition of APB SPI
 *
 * ******************************************************************************************
 */

// Includes ---------------------------------------------------------------------------------
#include "apb_spi_driver.h"


// Definitions ------------------------------------------------------------------------------

// For example APB SPI clock is 66MHz
#define APB_SPI_CLK    66000000

// APB SPI transfer operation
#define APB_SPI_SEND                      0x0
#define APB_SPI_RECEIVE                   0x1
#define APB_SPI_TRANSFER                  0x2


// Get APB SPI revision number
static inline unsigned int apb_spix_ip_revision_number(APB_SPI_RESOURCES *spi)
{
	return spi->reg->IDREV & (APB_SPI_IDREV_MAJOR_MASK | APB_SPI_IDREV_MINOR_MASK);
}

/* APB_SPI_REVISION_NUM(major, minor): Generate revision number from major and minor number. */
#define APB_SPI_REVISION_NUM(major, minor)  (((major << APB_SPI_IDREV_MAJOR_BIT) & APB_SPI_IDREV_MAJOR_MASK) | (minor & APB_SPI_IDREV_MINOR_MASK))

/*
 * APB SPI FIFO depth to 128, and add slave data-only transfer feature.
 */
#define APB_SPI_IP_HAS_FIFO_DEPTH_128(spi)  (apb_spix_ip_revision_number(spi) > APB_SPI_REVISION_NUM(4, 6))


// Get TX FIFO size
static unsigned int apb_spix_get_txfifo_size(APB_SPI_RESOURCES *spi)
{
	if (APB_SPI_IP_HAS_FIFO_DEPTH_128(spi))
	{
		return 2 << ((spi->reg->CONFIG >> 4) & 0xf);
	}
	else
	{
		return 2 << ((spi->reg->CONFIG >> 4) & 0x3);
	}
}

static void apb_spix_polling_spiactive(APB_SPI_RESOURCES *spi)
{
	while ((spi->reg->STATUS) & 0x1);
}

// Get capabilities
APB_SPI_CAPABILITIES apb_spix_get_capabilities(APB_SPI_RESOURCES *spi)
{
	return spi->capabilities;
}

// Initializes APB SPI
int apb_spix_initialize(APB_SPI_SignalEvent_t cb_event, APB_SPI_RESOURCES *spi)
{
	if (spi->info->flags & APB_SPI_FLAG_INITIALIZED)
	{
		return APB_SPI_DRIVER_OK;
	}

	// Initialize APB SPI run-time resources
	spi->info->cb_event          = cb_event;

	spi->info->status.busy       = 0U;
	spi->info->status.data_lost  = 0U;
	spi->info->status.mode_fault = 0U;

	spi->info->xfer.rx_buf       = 0U;
	spi->info->xfer.tx_buf       = 0U;
	spi->info->xfer.tx_buf_limit = 0U;
	spi->info->xfer.rx_cnt       = 0U;
	spi->info->xfer.tx_cnt       = 0U;

	spi->info->mode              = 0U;
	spi->info->record_rx_buf     = 0U;
	spi->info->is_header         = 0U;
	spi->info->txfifo_size 		 = apb_spix_get_txfifo_size(spi);

	spi->info->flags = APB_SPI_FLAG_INITIALIZED;  // APB SPI is initialized

	spi->info->xfer.dma_tx_complete = 0;
	spi->info->xfer.dma_rx_complete = 0;

	return APB_SPI_DRIVER_OK;
}

// Uninitialized
int apb_spix_uninitialize(APB_SPI_RESOURCES *spi)
{
	spi->info->flags = 0U;                    // APB SPI is uninitialized

	return APB_SPI_DRIVER_OK;
}

// Control
int apb_spix_power_control(APB_SPI_POWER_STATE state, APB_SPI_RESOURCES *spi)
{
	switch (state)
	{
		case APB_SPI_POWER_OFF:
			// Disable interrupt APB SPI source
			hal_disable_interrupt(spi->irq_num);

#if (APB_SPI_DMA_EN == 1)
			// DMA un-initialize
			if (spi->dma_tx || spi->dma_rx)
			{
				dma_uninitialize();

				if (spi->dma_tx && (spi->info->status.busy != 0))
				{
					dma_channel_disable(spi->dma_tx->channel);
				}

				if (spi->dma_rx && (spi->info->status.busy != 0))
				{
					dma_channel_disable(spi->dma_rx->channel);
				}
			}
#endif

			// Reset APB SPI and TX/RX FIFOs
			spi->reg->CTRL = (TXFIFORST | RXFIFORST | SPIRST);

			// Clear APB SPI run-time resources
			spi->info->status.busy       = 0U;
			spi->info->status.data_lost  = 0U;
			spi->info->status.mode_fault = 0U;

			spi->info->xfer.rx_buf       = 0U;
			spi->info->xfer.tx_buf       = 0U;
			spi->info->xfer.rx_cnt       = 0U;
			spi->info->xfer.tx_cnt       = 0U;

			spi->info->mode              = 0U;

			spi->info->flags &= ~APB_SPI_FLAG_POWERED;

			break;

		case APB_SPI_POWER_LOW:
			return APB_SPI_DRIVER_ERROR_UNSUPPORTED;

		case APB_SPI_POWER_FULL:
			if ((spi->info->flags & APB_SPI_FLAG_INITIALIZED) == 0U)
			{
				return APB_SPI_DRIVER_ERROR;
			}

			if ((spi->info->flags & APB_SPI_FLAG_POWERED) != 0U)
			{
				return APB_SPI_DRIVER_OK;
			}

#if (APB_SPI_DMA_EN == 1)
			// DMA initialize
			if (spi->dma_tx || spi->dma_rx)
			{
				dma_initialize();
			}
#endif

			// Reset APB SPI and TX/RX FIFOs
			spi->reg->CTRL = (TXFIFORST | RXFIFORST | SPIRST);

			// Clear APB SPI run-time resources
			spi->info->status.busy       = 0U;
			spi->info->status.data_lost  = 0U;
			spi->info->status.mode_fault = 0U;

			spi->info->xfer.rx_buf       = 0U;
			spi->info->xfer.tx_buf       = 0U;
			spi->info->xfer.rx_cnt       = 0U;
			spi->info->xfer.tx_cnt       = 0U;

			spi->info->mode              = 0U;

			// Enable APB SPI IRQ
			hal_enable_interrupt(spi->irq_num);

			spi->info->flags |= APB_SPI_FLAG_POWERED;   // APB SPI is powered

			break;

		default:
			return APB_SPI_DRIVER_ERROR_UNSUPPORTED;
	}

	return APB_SPI_DRIVER_OK;
}

// Send data
int apb_spix_send(const void *data, unsigned int num, APB_SPI_RESOURCES *spi)
{
#if (APB_SPI_DMA_EN == 1)
	int stat;
#endif

	if ((data == NULL) || (num == 0U))
	{
		return APB_SPI_DRIVER_ERROR_PARAMETER;
	}

	if (!(spi->info->flags & APB_SPI_FLAG_CONFIGURED))
	{
		return APB_SPI_DRIVER_ERROR;
	}

	if (spi->info->status.busy)
	{
		return APB_SPI_DRIVER_ERROR_BUSY;
	}

	// Set busy flag
	spi->info->status.busy       = 1U;

	spi->info->status.data_lost  = 0U;
	spi->info->status.mode_fault = 0U;

	spi->info->xfer.tx_buf   = (unsigned char *)data;
	spi->info->xfer.tx_cnt   = 0U;

	spi->info->xfer.transfer_op = APB_SPI_SEND;

	// Wait prior transfer finish
	apb_spix_polling_spiactive(spi);

	if ((spi->info->mode & APB_SPI_TRANSFER_FORMAT_Msk) != APB_SPI_SLV_DATA_ONLY_TRANSFER)
	{
		// Set transfer mode to write only and transfer count for write data
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSMODE_WRONLY | WR_TRANCNT(num));
	}
	else
	{
		// SlvDataOnly mode should set TransMode to 0
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSCTRL_SLV_DATA_ONLY | WR_TRANCNT(num));
	}

	// Set TX FIFO threshold to 2
	spi->reg->CTRL = TXTHRES(2);

	// Prepare info that ISR needed
	spi->info->xfer.txfifo_refill = spi->info->txfifo_size - 2; // TX FIFO threshold = 2

	if (spi->info->data_bits <= 8)
	{
		// Data bits = 1....8
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num));
		spi->info->src_width = 1;
	}
	else if (spi->info->data_bits <= 16)
	{
		// Data bits = 9....16
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num * 2));
		spi->info->src_width = 2;
	}
	else
	{
		// Data bits = 17....32
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num * 4));
		spi->info->src_width = 4;
	}

#if (APB_SPI_DMA_EN == 1)
	// DMA mode
	if (spi->dma_tx)
	{
		// Enable TX DMA
		spi->reg->CTRL |= TXDMAEN;

		// Configure DMA channel
		stat = dma_channel_configure(spi->dma_tx->channel,
					     (unsigned int)(long)spi->info->xfer.tx_buf,
					     (unsigned int)(long)(&(spi->reg->DATA)),
					     num,
					     DMA_CH_CTRL_SBINF(DMA_INF_IDX1) |
					     DMA_CH_CTRL_DBINF(DMA_INF_IDX1) |
					     DMA_CH_CTRL_SBSIZE(DMA_BSIZE_1) |
					     // DMA width setting: 1 byte:0x0, 2 byte:0x1, 4 byte:0x2
					     DMA_CH_CTRL_SWIDTH(spi->info->src_width/2) |
					     DMA_CH_CTRL_DWIDTH(spi->info->src_width/2) |
					     DMA_CH_CTRL_DMODE_HANDSHAKE |
					     DMA_CH_CTRL_SRCADDR_INC |
					     DMA_CH_CTRL_DSTADDR_FIX |
					     DMA_CH_CTRL_DSTREQ(spi->dma_tx->reqsel) |
					     DMA_CH_CTRL_INTABT |
					     DMA_CH_CTRL_INTERR |
					     DMA_CH_CTRL_INTTC |
					     DMA_CH_CTRL_ENABLE,
					     spi->dma_tx->cb_event);

		if (stat == -1)
		{
			return APB_SPI_DRIVER_ERROR;
		}

		// Enable interrupts
		spi->reg->INTREN = APB_SPI_ENDINT;
	}
	else
#endif
	{
		// Interrupt mode

		// Enable interrupts
		spi->reg->INTREN = (APB_SPI_TXFIFOINT | APB_SPI_ENDINT);

		// Enable TX FIFO under run interrupt when slave mode
		if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_SLAVE)
		{
			spi->reg->INTREN |= APB_SPI_TXFIFOOURINT;
		}
	}

	// Trigger transfer when APB SPI master mode
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_MASTER)
	{
		spi->reg->CMD = 0U;
	}

	return APB_SPI_DRIVER_OK;
}

// Receive data
int apb_spix_receive(void *data, unsigned int num, APB_SPI_RESOURCES *spi)
{
#if (APB_SPI_DMA_EN == 1)
	int stat;
#endif

	if ((data == NULL) || (num == 0U))
	{
		return APB_SPI_DRIVER_ERROR_PARAMETER;
	}

	if (!(spi->info->flags & APB_SPI_FLAG_CONFIGURED))
	{
		return APB_SPI_DRIVER_ERROR;
	}

	if (spi->info->status.busy)
	{
		return APB_SPI_DRIVER_ERROR_BUSY;
	}

	// Set busy flag
	spi->info->status.busy       = 1U;

	spi->info->status.data_lost  = 0U;
	spi->info->status.mode_fault = 0U;

	spi->info->xfer.rx_buf = (unsigned char *)data;
	spi->info->xfer.rx_cnt = 0U;

	spi->info->xfer.transfer_op = APB_SPI_RECEIVE;

	// Wait prior transfer finish
	apb_spix_polling_spiactive(spi);

	if ((spi->info->mode & APB_SPI_TRANSFER_FORMAT_Msk) != APB_SPI_SLV_DATA_ONLY_TRANSFER)
	{
		// Set transfer mode to read only and transfer count for read data
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSMODE_RDONLY | RD_TRANCNT(num));
	}
	else
	{
		// SlvDataOnly mode should set TransMode to 0
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSCTRL_SLV_DATA_ONLY | RD_TRANCNT(num));
	}

	// Set RX FIFO threshold to 2
	spi->reg->CTRL = RXTHRES(1);

	if (spi->info->data_bits <= 8)
	{
		// Data bits = 1....8
		spi->info->src_width = 1;
	}
	else if (spi->info->data_bits <= 16)
	{
		// Data bits = 9....16
		spi->info->src_width = 2;
	}
	else
	{
		// Data bits = 17....32
		spi->info->src_width = 4;
	}

#if (APB_SPI_DMA_EN == 1)
	// DMA mode
	if (spi->dma_rx)
	{
		// Enable RX DMA
		spi->reg->CTRL |= RXDMAEN;

		// Configure DMA channel
		stat = dma_channel_configure(spi->dma_rx->channel,
					     (unsigned int)(long)(&(spi->reg->DATA)),
					     (unsigned int)(long)spi->info->xfer.rx_buf,
					     num,
					     DMA_CH_CTRL_SBINF(DMA_INF_IDX1) |
					     DMA_CH_CTRL_DBINF(DMA_INF_IDX1) |
					     DMA_CH_CTRL_SBSIZE(DMA_BSIZE_1) |
					     // DMA width setting: 1 byte:0x0, 2 byte:0x1, 4 byte:0x2
					     DMA_CH_CTRL_SWIDTH(spi->info->src_width/2) |
					     DMA_CH_CTRL_DWIDTH(spi->info->src_width/2) |
					     DMA_CH_CTRL_SMODE_HANDSHAKE |
					     DMA_CH_CTRL_SRCADDR_FIX |
					     DMA_CH_CTRL_DSTADDR_INC |
					     DMA_CH_CTRL_SRCREQ(spi->dma_rx->reqsel) |
					     DMA_CH_CTRL_INTABT |
					     DMA_CH_CTRL_INTERR |
					     DMA_CH_CTRL_INTTC |
					     DMA_CH_CTRL_ENABLE,
					     spi->dma_rx->cb_event);

		if (stat == -1)
		{
			return APB_SPI_DRIVER_ERROR;
		}

		// Enable interrupts
		spi->reg->INTREN = APB_SPI_ENDINT;
	}
	else
#endif
	{
		// Interrupt mode

		// Enable interrupts
		spi->reg->INTREN = (APB_SPI_RXFIFOINT | APB_SPI_ENDINT);

		// Enable RX FIFO overrun interrupt when slave mode
		if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_SLAVE)
		{
			spi->reg->INTREN |= APB_SPI_RXFIFOOORINT;
		}
	}

	// Trigger transfer when APB_SPI master mode
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_MASTER)
	{
		spi->reg->CMD = 0U;
	}

	return APB_SPI_DRIVER_OK;
}

// Transfer data
int apb_spix_transfer(const void *data_out, void *data_in, unsigned int num, APB_SPI_RESOURCES *spi)
{
	unsigned int dma_rx_num = 0;

#if (APB_SPI_DMA_EN == 1)
	int stat;
#endif

	if ((data_out == NULL) || (data_in == NULL) || (num == 0U))
	{
		return APB_SPI_DRIVER_ERROR_PARAMETER;
	}

	if (!(spi->info->flags & APB_SPI_FLAG_CONFIGURED))
	{
		return APB_SPI_DRIVER_ERROR;
	}

	if (spi->info->status.busy)
	{
		return APB_SPI_DRIVER_ERROR_BUSY;
	}

	if (spi->info->tx_header_len > num)
	{
		return APB_SPI_ERROR_HEADER_LEN;
	}

	// Set busy flag
	spi->info->status.busy       = 1U;

	spi->info->status.data_lost  = 0U;
	spi->info->status.mode_fault = 0U;

	spi->info->xfer.rx_buf = (unsigned char *)data_in;
	spi->info->xfer.tx_buf = (unsigned char *)data_out;
	spi->info->xfer.rx_cnt = 0U;
	spi->info->xfer.tx_cnt = 0U;

	spi->info->xfer.transfer_op = APB_SPI_TRANSFER;

	// Wait prior transfer finish
	apb_spix_polling_spiactive(spi);

	if ((spi->info->mode & APB_SPI_TRANSFER_FORMAT_Msk) != APB_SPI_SLV_DATA_ONLY_TRANSFER)
	{
		// Set transfer mode to write and read at the same time and transfer count for write/read data
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSMODE_WRnRD | WR_TRANCNT(num) | RD_TRANCNT(num));
	}
	else
	{
		// SlvDataOnly mode should set TransMode to 0
		spi->reg->TRANSCTRL &= ~(APB_SPI_TRANSMODE_MSK | RD_TRANCNT_MSK | WR_TRANCNT_MSK);
		spi->reg->TRANSCTRL |= (APB_SPI_TRANSCTRL_SLV_DATA_ONLY | WR_TRANCNT(num) | RD_TRANCNT(num));
	}

	if (spi->info->data_bits <= 8)
	{
		// Data bits = 1....8
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num));
		spi->info->src_width = 1;
	}
	else if (spi->info->data_bits <= 16)
	{
		// Data bits = 9....16
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num * 2));
		spi->info->src_width = 2;
	}
	else
	{
		// Data bits = 17....32
		spi->info->xfer.tx_buf_limit = (unsigned char *)((long)(spi->info->xfer.tx_buf + num * 4));
		spi->info->src_width = 4;
	}

	if ((spi->info->tx_header_len > 0) && (spi->info->mode & APB_SPI_CONTROL_Msk))
	{
		// If there is TX header, only setting DMA to receive the header at first time
		dma_rx_num = spi->info->tx_header_len;
		spi->info->record_rx_buf = (unsigned char *)(spi->info->xfer.rx_buf);
		spi->info->is_header = 1;
		spi->info->data_num = num - spi->info->tx_header_len;
	}
	else
	{
		dma_rx_num = num;
	}

	spi->reg->CTRL &= 0x0;

	// Set TX FIFO threshold and RX FIFO threshold to 2
	spi->reg->CTRL |= (TXTHRES(2) | RXTHRES(1));

	// Prepare info that ISR needed
	spi->info->xfer.txfifo_refill = spi->info->txfifo_size - 2; // TX FIFO threshold = 2

#if (APB_SPI_DMA_EN == 1)
	// DMA mode
	if (spi->dma_tx || spi->dma_rx)
	{
		if (spi->dma_tx)
		{
			// Enable TX DMA
			spi->reg->CTRL |= TXDMAEN;

			// Configure DMA channel
			stat = dma_channel_configure(spi->dma_tx->channel,
						      (unsigned int)(long)spi->info->xfer.tx_buf,
						      (unsigned int)(long)(&(spi->reg->DATA)),
						      num,
						      DMA_CH_CTRL_SBINF(DMA_INF_IDX1) |
						      DMA_CH_CTRL_DBINF(DMA_INF_IDX1) |
						      DMA_CH_CTRL_SBSIZE(DMA_BSIZE_1) |
						      // DMA width setting: 1 byte:0x0, 2 byte:0x1, 4 byte:0x2
						      DMA_CH_CTRL_SWIDTH(spi->info->src_width/2) |
						      DMA_CH_CTRL_DWIDTH(spi->info->src_width/2) |
						      DMA_CH_CTRL_DMODE_HANDSHAKE |
						      DMA_CH_CTRL_SRCADDR_INC |
						      DMA_CH_CTRL_DSTADDR_FIX |
						      DMA_CH_CTRL_DSTREQ(spi->dma_tx->reqsel) |
						      DMA_CH_CTRL_INTABT |
						      DMA_CH_CTRL_INTERR |
						      DMA_CH_CTRL_INTTC |
						      DMA_CH_CTRL_ENABLE,
						      spi->dma_tx->cb_event);
			if (stat == -1)
			{
				return APB_SPI_DRIVER_ERROR;
			}
		}

		if (spi->dma_rx)
		{
			// Enable RX DMA
			spi->reg->CTRL |= RXDMAEN;

			// Configure DMA channel
			stat = dma_channel_configure(spi->dma_rx->channel,
						     (unsigned int)(long)(&(spi->reg->DATA)),
						     (unsigned int)(long)spi->info->xfer.rx_buf,
						     dma_rx_num,
						     DMA_CH_CTRL_SBINF(DMA_INF_IDX0) |
						     DMA_CH_CTRL_DBINF(DMA_INF_IDX0) |
						     DMA_CH_CTRL_SBSIZE(DMA_BSIZE_1) |
						     // DMA width setting: 1 byte:0x0, 2 byte:0x1, 4 byte:0x2
						     DMA_CH_CTRL_SWIDTH(spi->info->src_width/2) |
						     DMA_CH_CTRL_DWIDTH(spi->info->src_width/2) |
						     DMA_CH_CTRL_SMODE_HANDSHAKE |
						     DMA_CH_CTRL_SRCADDR_FIX |
						     DMA_CH_CTRL_DSTADDR_INC |
						     DMA_CH_CTRL_SRCREQ(spi->dma_rx->reqsel) |
						     DMA_CH_CTRL_INTABT |
						     DMA_CH_CTRL_INTERR |
						     DMA_CH_CTRL_INTTC |
						     DMA_CH_CTRL_ENABLE,
						     spi->dma_rx->cb_event);
			if (stat == -1)
			{
				return APB_SPI_DRIVER_ERROR;
			}
		}

		// Enable interrupts
		spi->reg->INTREN = APB_SPI_ENDINT;
	}
	else
#endif
	{
		// Interrupt mode

		// Enable interrupts
		spi->reg->INTREN = (APB_SPI_TXFIFOINT | APB_SPI_RXFIFOINT | APB_SPI_ENDINT);

		// Enable TX FIFO under run and RX FIFO overrun interrupt when slave mode
		if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_SLAVE)
		{
			spi->reg->INTREN |= (APB_SPI_TXFIFOOURINT | APB_SPI_RXFIFOOORINT);
		}
	}

	// Trigger transfer when APB SPI master mode
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_MASTER)
	{
		spi->reg->CMD = 0;
	}

	return APB_SPI_DRIVER_OK;
}

// Get data count
int apb_spix_get_data_count(APB_SPI_RESOURCES *spi)
{
	if (!(spi->info->flags & APB_SPI_FLAG_CONFIGURED))
	{
		return 0U;
	}

	switch (spi->info->xfer.transfer_op)
	{
		case APB_SPI_SEND:
		case APB_SPI_TRANSFER:
#if (APB_SPI_DMA_EN == 1)
			if (spi->dma_tx)
			{
				return (dma_channel_get_count(spi->dma_tx->channel));
			}
			else
#endif
			{
				return (spi->info->xfer.tx_cnt);
			}
		case APB_SPI_RECEIVE:
#if (APB_SPI_DMA_EN == 1)
			if (spi->dma_rx)
			{
				return (dma_channel_get_count(spi->dma_rx->channel));
			}
			else
#endif
			{
				return (spi->info->xfer.rx_cnt);
			}
		default:
			return APB_SPI_DRIVER_OK;
	}
}

// Control
int apb_spix_control(unsigned int control, unsigned int arg, APB_SPI_RESOURCES *spi)
{
	unsigned int sclk_div;

	if (!(spi->info->flags & APB_SPI_FLAG_POWERED))
	{
		return APB_SPI_DRIVER_ERROR;
	}

	if ((control & APB_SPI_CONTROL_Msk) == APB_SPI_ABORT_TRANSFER)
	{
		// Abort APB SPI transfer
		// Disable APB SPI interrupts
		spi->reg->INTREN = 0;

		// Clear APB SPI run-time resources
		spi->info->status.busy = 0U;

		spi->info->xfer.rx_buf = 0U;
		spi->info->xfer.tx_buf = 0U;
		spi->info->xfer.rx_cnt = 0U;
		spi->info->xfer.tx_cnt = 0U;

		spi->info->mode        = 0U;

		return APB_SPI_DRIVER_OK;
	}

	if (spi->info->status.busy)
	{
		return APB_SPI_DRIVER_ERROR_BUSY;
	}

	switch (control & APB_SPI_CONTROL_Msk)
	{
		case APB_SPI_MODE_MASTER_SIMPLEX:  // APB SPI master (output/input on MOSI); arg = bus speed in bps
		case APB_SPI_MODE_SLAVE_SIMPLEX:   // APB SPI slave (output/input on MISO
		case APB_SPI_SET_DEFAULT_TX_VALUE: // Set default Transmit value; arg = value
		case APB_SPI_CONTROL_SS:           // Control slave select; arg = 0:inactive, 1:active
			return APB_SPI_ERROR_MODE;

		case APB_SPI_MODE_INACTIVE:		 // APB SPI inactive
			spi->info->mode &= ~APB_SPI_CONTROL_Msk;
			spi->info->mode |= APB_SPI_MODE_INACTIVE;

			spi->info->flags &= ~APB_SPI_FLAG_CONFIGURED;

			return APB_SPI_DRIVER_OK;

		case APB_SPI_MODE_MASTER:
			// APB SPI master (output on MOSI, input on MISO); arg = bus speed in bps
			// Set master mode and disable data merge mode
			spi->reg->TRANSFMT &= ~(APB_SPI_MERGE | APB_SPI_SLAVE);

			spi->info->mode &= ~APB_SPI_CONTROL_Msk;
			spi->info->mode |= APB_SPI_MODE_MASTER;

			spi->info->flags |= APB_SPI_FLAG_CONFIGURED;

			goto set_speed;

		case APB_SPI_MODE_SLAVE:
			// APB SPI slave (output on MISO, input on MOSI)
			// Set slave mode and disable data merge mode
			spi->reg->TRANSFMT &= ~APB_SPI_MERGE;
			spi->reg->TRANSFMT |= APB_SPI_SLAVE;

			spi->info->mode &= ~APB_SPI_CONTROL_Msk;
			spi->info->mode |= APB_SPI_MODE_SLAVE;

			spi->info->flags |= APB_SPI_FLAG_CONFIGURED;

			break;

		case APB_SPI_SET_BUS_SPEED:
			// Set bus speed in bps; arg = value
			set_speed:
				if (arg == 0U)
				{
					return APB_SPI_DRIVER_ERROR;
				}

				sclk_div = (APB_SPI_CLK / (2 * arg)) - 1;

				spi->reg->TIMING &= ~0xff;
				spi->reg->TIMING |= sclk_div;

				if ((control & APB_SPI_CONTROL_Msk) == APB_SPI_SET_BUS_SPEED)
				{
					return APB_SPI_DRIVER_OK;
				}

			break;

		case APB_SPI_GET_BUS_SPEED:
			// Get bus speed in bps
			sclk_div = spi->reg->TIMING & 0xff;
			return (APB_SPI_CLK / ((sclk_div + 1) * 2));

		case APB_SPI_TX_HEADER_LENGTH:
			spi->info->tx_header_len = (unsigned char)arg;
			return APB_SPI_DRIVER_OK;

		default:
			return APB_SPI_DRIVER_ERROR_UNSUPPORTED;
	}

	// APB SPI slave select mode for master
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) ==  APB_SPI_MODE_MASTER)
	{
		switch (control & APB_SPI_SS_MASTER_MODE_Msk)
		{
			case APB_SPI_SS_MASTER_HW_OUTPUT:
				break;
			case APB_SPI_SS_MASTER_UNUSED:
			case APB_SPI_SS_MASTER_SW:
			case APB_SPI_SS_MASTER_HW_INPUT:
				return APB_SPI_ERROR_SS_MODE;
			default:
				break;
		}
	}

	// APB SPI slave select mode for slave
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_SLAVE)
	{
		switch (control & APB_SPI_SS_SLAVE_MODE_Msk)
		{
			case APB_SPI_SS_SLAVE_HW:
				break;
			case APB_SPI_SS_SLAVE_SW:
				return APB_SPI_ERROR_SS_MODE;
			default:
				break;
		}
	}

	// APB SPI data-only transfer mode for slave
	if ((spi->info->mode & APB_SPI_CONTROL_Msk) == APB_SPI_MODE_SLAVE)
	{
		if ((control & APB_SPI_TRANSFER_FORMAT_Msk) == APB_SPI_SLV_DATA_ONLY_TRANSFER)
		{
			spi->info->mode &= ~APB_SPI_TRANSFER_FORMAT_Msk;
			spi->info->mode |= APB_SPI_SLV_DATA_ONLY_TRANSFER;
		}
	}

	// Set APB SPI frame format
	switch (control & APB_SPI_FRAME_FORMAT_Msk)
	{
		case APB_SPI_CPOL0_CPHA0:
			spi->reg->TRANSFMT &= ~(APB_SPI_CPOL | APB_SPI_CPHA);
			break;
		case APB_SPI_CPOL0_CPHA1:
			spi->reg->TRANSFMT &= ~APB_SPI_CPOL;
			spi->reg->TRANSFMT |= APB_SPI_CPHA;
			break;
		case APB_SPI_CPOL1_CPHA0:
			spi->reg->TRANSFMT |= APB_SPI_CPOL;
			spi->reg->TRANSFMT &= ~APB_SPI_CPHA;
			break;
		case APB_SPI_CPOL1_CPHA1:
			spi->reg->TRANSFMT |= APB_SPI_CPOL;
			spi->reg->TRANSFMT |= APB_SPI_CPHA;
			break;
		case APB_SPI_TI_SSI:
		case APB_SPI_MICROWIRE:
		default:
			return APB_SPI_ERROR_FRAME_FORMAT;
	}

	// Set number of data bits
	spi->info->data_bits = ((control & APB_SPI_DATA_BITS_Msk) >> APB_SPI_DATA_BITS_Pos);

	if ((spi->info->data_bits < 1U) || (spi->info->data_bits > 32U))
	{
		return APB_SPI_ERROR_DATA_BITS;
	}
	else
	{
		spi->reg->TRANSFMT &= ~(DATA_BITS_MSK);
		spi->reg->TRANSFMT |= DATA_BITS(spi->info->data_bits);
	}

	// Set APB SPI bit order
	if ((control & APB_SPI_BIT_ORDER_Msk) == APB_SPI_LSB_MSB)
	{
		spi->reg->TRANSFMT |= APB_SPI_LSB;
	}
	else
	{
		spi->reg->TRANSFMT &= ~APB_SPI_LSB;
	}

	return APB_SPI_DRIVER_OK;
}

// Get status
APB_SPI_STATUS apb_spix_get_status(APB_SPI_RESOURCES *spi)
{
	APB_SPI_STATUS status;

	status.busy       = spi->info->status.busy;
	status.data_lost  = spi->info->status.data_lost;
	status.mode_fault = spi->info->status.mode_fault;

	return (status);
}

// Interrupt handler
void apb_spix_irq_handler(APB_SPI_RESOURCES *spi)
{
	unsigned int i, j, status;
	unsigned int data = 0;
	unsigned int rx_num = 0;
	unsigned int event = 0;

	// Read status register
	status = spi->reg->INTRST;

	if ((status & APB_SPI_RXFIFOOORINT) || (status & APB_SPI_TXFIFOOURINT))
	{
		// TX FIFO under run or RX FIFO overrun interrupt status
		spi->info->status.data_lost = 1U;

		event |= APB_SPI_EVENT_DATA_LOST;
	}

	if (status & APB_SPI_TXFIFOINT)
	{
		for (i = 0; i < spi->info->xfer.txfifo_refill; i++)
		{
			data = 0;
			if (spi->info->xfer.tx_buf < spi->info->xfer.tx_buf_limit)
			{
				// Handle the data frame format
				if (spi->info->data_bits <= 8)
				{
					// Data bits = 1....8
					data = *spi->info->xfer.tx_buf;
					spi->info->xfer.tx_buf++;
				}
				else if (spi->info->data_bits <= 16)
				{
					// Data bits = 9....16
					for (j = 0; j < 2; j++)
					{
						data |= *spi->info->xfer.tx_buf << j * 8;
						spi->info->xfer.tx_buf++;
					}
				}
				else
				{
					// Data bits = 17....32
					for (j = 0; j < 4; j++)
					{
						data |= *spi->info->xfer.tx_buf << j * 8;
						spi->info->xfer.tx_buf++;
					}
				}

				spi->reg->DATA = data;
				spi->info->xfer.tx_cnt++;
			}
			else
			{
				spi->reg->INTREN &= ~APB_SPI_TXFIFOINT;
			}
		}
	}

	if (status & APB_SPI_RXFIFOINT)
	{
		// Get number of valid entries in the RX FIFO
		if (APB_SPI_IP_HAS_FIFO_DEPTH_128(spi))
		{
			rx_num = (((spi->reg->STATUS >> 8) & 0x2f) | ((spi->reg->STATUS & (0x3 << 24)) >> 18));
		}
		else
		{
			rx_num = (spi->reg->STATUS >> 8) & 0x1f;
		}

		for (i = rx_num; i > 0; i--)
		{
			data = spi->reg->DATA;
			spi->info->xfer.rx_cnt++;

			// Handle the data frame format
			if (spi->info->data_bits <= 8)
			{
				*spi->info->xfer.rx_buf = data & 0xff;
				spi->info->xfer.rx_buf++;
			}
			else if (spi->info->data_bits <= 16)
			{
				for (j = 0; j < 2; j++)
				{
					*spi->info->xfer.rx_buf = (data >> j * 8) & 0xff;
					spi->info->xfer.rx_buf++;
				}
			}
			else
			{
				for (j = 0; j < 4; j++)
				{
					*spi->info->xfer.rx_buf = (data >> j * 8) & 0xff;
					spi->info->xfer.rx_buf++;
				}
			}

			// If there is a TX header on spi_transfer, drop the dummy data from slave when master sending header.
			if ((spi->info->is_header == 1) && (spi->info->xfer.rx_cnt == spi->info->tx_header_len))
			{
				spi->info->is_header = 0;
				spi->info->xfer.rx_buf = spi->info->record_rx_buf;
				spi->info->xfer.rx_cnt = 0;
			}
		}
	}

	if (status & APB_SPI_ENDINT)
	{
		// Disable APB_SPI interrupts
		spi->reg->INTREN = 0;

#if (APB_SPI_DMA_EN == 1)
		if (spi->dma_tx && ((spi->info->xfer.transfer_op == APB_SPI_SEND) || (spi->info->xfer.transfer_op == APB_SPI_TRANSFER)))
		{
			if (!spi->info->xfer.dma_tx_complete)
			{
				dma_channel_disable(spi->dma_tx->channel);
			}

			spi->info->xfer.dma_tx_complete = 0;
		}

		if (spi->dma_rx && ((spi->info->xfer.transfer_op == APB_SPI_RECEIVE) || (spi->info->xfer.transfer_op == APB_SPI_TRANSFER)))
		{
			if (!spi->info->xfer.dma_rx_complete)
			{
				dma_channel_disable(spi->dma_rx->channel);
			}

			spi->info->xfer.dma_rx_complete = 0;
		}
#endif

		// Clear TX/RX FIFOs
		spi->reg->CTRL = (TXFIFORST | RXFIFORST);

		spi->info->status.busy = 0U;

		event |= APB_SPI_EVENT_TRANSFER_COMPLETE;
	}

	// Clear interrupt status
	spi->reg->INTRST = status;
	// Make sure "write 1 clear" take effect before return
	spi->reg->INTRST;

	if ((spi->info->cb_event != NULL) && (event != 0))
	{
		spi->info->cb_event(event);
	}
}

#if (APB_SPI_DMA_EN == 1)
// DMA TX event
void apb_spix_dma_tx_event (unsigned int event, APB_SPI_RESOURCES * spi)
{
	switch (event)
	{
		case DMA_EVENT_TERMINAL_COUNT_REQUEST:
			spi->info->xfer.dma_tx_complete = 1;
			break;
		case DMA_EVENT_ERROR:
		default:
			break;
	}
}

// DMA RX event
void apb_spix_dma_rx_event (unsigned int event, APB_SPI_RESOURCES * spi)
{
	switch (event)
	{
		case DMA_EVENT_TERMINAL_COUNT_REQUEST:
			if (spi->info->is_header == 0)
			{
				spi->info->xfer.dma_rx_complete = 1;
			}
			else
			{
				// Setting another DMA transfer to cover the dummy data from slave when master is sending header.
				// Configure DMA channel
				dma_channel_configure(spi->dma_rx->channel,
						       (unsigned int)(long)(&(spi->reg->DATA)),
						       (unsigned int)(long)spi->info->xfer.rx_buf,
						       spi->info->data_num,
						       DMA_CH_CTRL_SBINF(DMA_INF_IDX0) |
						       DMA_CH_CTRL_DBINF(DMA_INF_IDX0) |
						       DMA_CH_CTRL_SBSIZE(DMA_BSIZE_1) |
						       // DMA width setting: 1 byte:0x0, 2 byte:0x1, 4 byte:0x2
						       DMA_CH_CTRL_SWIDTH(spi->info->src_width/2) |
						       DMA_CH_CTRL_DWIDTH(spi->info->src_width/2) |
						       DMA_CH_CTRL_SMODE_HANDSHAKE |
						       DMA_CH_CTRL_SRCADDR_FIX |
						       DMA_CH_CTRL_DSTADDR_INC |
						       DMA_CH_CTRL_SRCREQ(spi->dma_rx->reqsel) |
						       DMA_CH_CTRL_INTABT |
						       DMA_CH_CTRL_INTERR |
						       DMA_CH_CTRL_INTTC |
						       DMA_CH_CTRL_ENABLE,
						       spi->dma_rx->cb_event);
				spi->info->is_header = 0;
			}

			break;

		case DMA_EVENT_ERROR:
			break;

		default:
			break;
	}
}
#endif
