/*
 * ******************************************************************************************
 *
 * File		: apb_spi_driver.h
 * Author	: GOWIN
 * Function	: Driver definition of APB SPI
 *
 * ******************************************************************************************
 */

#ifndef __APB_SPI_DRIVER_H__
#define __APB_SPI_DRIVER_H__


#ifdef __cplusplus
extern "C" {
#endif


// Includes ---------------------------------------------------------------------------------
#include "apb_spi.h"
#include "apb_spi_hal.h"
#include "apb_spi_hw.h"


// Definitions ------------------------------------------------------------------------------

/****** APB SPI Control Codes *****/

#define APB_SPI_CONTROL_Pos              0
#define APB_SPI_CONTROL_Msk             (0xFFUL << APB_SPI_CONTROL_Pos)

/*----- APB SPI Control Codes: Mode -----*/
#define APB_SPI_MODE_INACTIVE           (0x00UL << APB_SPI_CONTROL_Pos)      	// APB SPI inactive
#define APB_SPI_MODE_MASTER             (0x01UL << APB_SPI_CONTROL_Pos)      	// APB SPI Master (output on MOSI, input on MISO); arg = bus speed in bps
#define APB_SPI_MODE_SLAVE              (0x02UL << APB_SPI_CONTROL_Pos)      	// APB SPI Slave  (output on MISO, input on MOSI)
#define APB_SPI_MODE_MASTER_SIMPLEX     (0x03UL << APB_SPI_CONTROL_Pos)      	// APB SPI Master (output/input on MOSI); arg = bus speed in bps
#define APB_SPI_MODE_SLAVE_SIMPLEX      (0x04UL << APB_SPI_CONTROL_Pos)      	// APB SPI Slave  (output/input on MISO)

/*----- APB SPI Control Codes: Mode Parameters: Frame Format -----*/
#define APB_SPI_FRAME_FORMAT_Pos         8
#define APB_SPI_FRAME_FORMAT_Msk        (7UL << APB_SPI_FRAME_FORMAT_Pos)
#define APB_SPI_CPOL0_CPHA0             (0UL << APB_SPI_FRAME_FORMAT_Pos)    	// Clock polarity 0, clock phase 0 (default)
#define APB_SPI_CPOL0_CPHA1             (1UL << APB_SPI_FRAME_FORMAT_Pos)    	// Clock polarity 0, clock phase 1
#define APB_SPI_CPOL1_CPHA0             (2UL << APB_SPI_FRAME_FORMAT_Pos)    	// Clock polarity 1, clock phase 0
#define APB_SPI_CPOL1_CPHA1             (3UL << APB_SPI_FRAME_FORMAT_Pos)    	// Clock polarity 1, clock phase 1
#define APB_SPI_TI_SSI                  (4UL << APB_SPI_FRAME_FORMAT_Pos)    	// Texas instruments frame format
#define APB_SPI_MICROWIRE               (5UL << APB_SPI_FRAME_FORMAT_Pos)    	// National microwire frame format

/*----- APB SPI Control Codes: Mode Parameters: Data Bits -----*/
#define APB_SPI_DATA_BITS_Pos            12
#define APB_SPI_DATA_BITS_Msk           (0x3FUL << APB_SPI_DATA_BITS_Pos)
#define APB_SPI_DATA_BITS(n)            (((n) & 0x3F) << APB_SPI_DATA_BITS_Pos) // Number of data bits

/*----- APB SPI Control Codes: Mode Parameters: Bit Order -----*/
#define APB_SPI_BIT_ORDER_Pos            18
#define APB_SPI_BIT_ORDER_Msk           (1UL << APB_SPI_BIT_ORDER_Pos)
#define APB_SPI_MSB_LSB                 (0UL << APB_SPI_BIT_ORDER_Pos)      	// APB SPI bit order from MSB to LSB (default)
#define APB_SPI_LSB_MSB                 (1UL << APB_SPI_BIT_ORDER_Pos)     	 	// APB SPI bit order from LSB to MSB

/*----- APB SPI Control Codes: Mode Parameters: Slave Select Mode -----*/
#define APB_SPI_SS_MASTER_MODE_Pos       19
#define APB_SPI_SS_MASTER_MODE_Msk      (3UL << APB_SPI_SS_MASTER_MODE_Pos)
#define APB_SPI_SS_MASTER_UNUSED        (1UL << APB_SPI_SS_MASTER_MODE_Pos) 	// APB SPI Slave select when Master: not used
#define APB_SPI_SS_MASTER_SW            (2UL << APB_SPI_SS_MASTER_MODE_Pos) 	// APB SPI Slave select when Master: software controlled
#define APB_SPI_SS_MASTER_HW_OUTPUT     (0UL << APB_SPI_SS_MASTER_MODE_Pos) 	// APB SPI Slave select when Master: hardware controlled output (default)
#define APB_SPI_SS_MASTER_HW_INPUT      (3UL << APB_SPI_SS_MASTER_MODE_Pos) 	// APB SPI Slave select when Master: hardware monitored input

#define APB_SPI_SS_SLAVE_MODE_Pos        21
#define APB_SPI_SS_SLAVE_MODE_Msk       (1UL << APB_SPI_SS_SLAVE_MODE_Pos)
#define APB_SPI_SS_SLAVE_HW             (0UL << APB_SPI_SS_SLAVE_MODE_Pos)  	// APB SPI Slave select when Slave: hardware monitored (default)
#define APB_SPI_SS_SLAVE_SW             (1UL << APB_SPI_SS_SLAVE_MODE_Pos)  	// APB SPI Slave select when Slave: software controlled

/*----- APB SPI Control Codes: Mode Parameters: Transfer Format Phase -----*/
#define APB_SPI_TRANSFER_FORMAT_Pos      22
#define APB_SPI_TRANSFER_FORMAT_Msk     (1UL << APB_SPI_TRANSFER_FORMAT_Pos)
#define APB_SPI_SLV_DATA_ONLY_TRANSFER  (1UL << APB_SPI_TRANSFER_FORMAT_Pos)	// (Slave mode only) APB SPI transaction only consists of data without command and address phase.

/*----- APB SPI Control Codes: Miscellaneous Controls  -----*/
#define APB_SPI_SET_BUS_SPEED           (0x10UL << APB_SPI_CONTROL_Pos)     	// Set bus speed in bps; arg = value
#define APB_SPI_GET_BUS_SPEED           (0x11UL << APB_SPI_CONTROL_Pos)     	// Get bus speed in bps
#define APB_SPI_SET_DEFAULT_TX_VALUE    (0x12UL << APB_SPI_CONTROL_Pos)     	// Set default transmit value; arg = value
#define APB_SPI_CONTROL_SS              (0x13UL << APB_SPI_CONTROL_Pos)     	// Control Slave select; arg: 0=inactive, 1=active
#define APB_SPI_ABORT_TRANSFER          (0x14UL << APB_SPI_CONTROL_Pos)     	// Abort current data transfer
#define APB_SPI_TX_HEADER_LENGTH		(0x20UL << APB_SPI_CONTROL_Pos)

/****** APB SPI Slave Select Signal definitions *****/
#define APB_SPI_SS_INACTIVE              0                                    	// APB SPI Slave select signal inactive
#define APB_SPI_SS_ACTIVE                1                                    	// APB SPI Slave select signal active

/****** APB SPI specific error codes *****/
#define APB_SPI_ERROR_MODE              (APB_SPI_DRIVER_ERROR_SPECIFIC - 1)     // Specified mode not supported
#define APB_SPI_ERROR_FRAME_FORMAT      (APB_SPI_DRIVER_ERROR_SPECIFIC - 2)     // Specified frame format not supported
#define APB_SPI_ERROR_DATA_BITS         (APB_SPI_DRIVER_ERROR_SPECIFIC - 3)     // Specified number of data bits not supported
#define APB_SPI_ERROR_BIT_ORDER         (APB_SPI_DRIVER_ERROR_SPECIFIC - 4)     // Specified bit order not supported
#define APB_SPI_ERROR_SS_MODE           (APB_SPI_DRIVER_ERROR_SPECIFIC - 5)     // Specified Slave select mode not supported
#define APB_SPI_ERROR_HEADER_LEN		(APB_SPI_DRIVER_ERROR_SPECIFIC - 6)


/****** APB SPI Event *****/
#define APB_SPI_EVENT_TRANSFER_COMPLETE (1UL << 0)  // Data transfer completed
#define APB_SPI_EVENT_DATA_LOST         (1UL << 1)  // Data lost: Receive overflow / Transmit underflow
#define APB_SPI_EVENT_MODE_FAULT        (1UL << 2)  // Master mode fault (SS deactivated when Master)



/******************************************************************************
  \fn          void APB_SPI_SignalEvent (unsigned int event)
  \brief       Signal APB SPI Events.
  \param[in]   event \ref APB_SPI_events notification mask
  \return      none
*******************************************************************************/
typedef void (*APB_SPI_SignalEvent_t) (unsigned int event);  // Pointer to \ref APB_SPI_SignalEvent : Signal APB SPI Event.


/******************************************************************************
\brief APB SPI Driver Capabilities.
*******************************************************************************/
typedef struct _APB_SPI_CAPABILITIES
{
	unsigned int simplex          : 1;        // Supports simplex mode (Master and Slave)
	unsigned int ti_ssi           : 1;        // Supports TI synchronous serial interface
	unsigned int microwire        : 1;        // Supports microwire interface
	unsigned int event_mode_fault : 1;        // Signal mode fault event: \ref APB_SPI_EVENT_MODE_FAULT
} APB_SPI_CAPABILITIES;


/* APB SPI ID revision register */
#define APB_SPI_IDREV_MINOR_BIT           (0)
#define APB_SPI_IDREV_MINOR_MASK          (0xf << 0)
#define APB_SPI_IDREV_MAJOR_BIT           (4)
#define APB_SPI_IDREV_MAJOR_MASK          (0xf << 4)

/* APB SPI transfer format register */
#define APB_SPI_CPHA                      (1UL << 0)
#define APB_SPI_CPOL                      (1UL << 1)
#define APB_SPI_SLAVE                     (1UL << 2)
#define APB_SPI_LSB                       (1UL << 3)
#define APB_SPI_MERGE                     (1UL << 7)
#define DATA_BITS_MSK                 	  (0x1f << 8)
#define DATA_BITS(data_bits)              ((data_bits - 1) << 8)

/* APB SPI transfer control register */
#define APB_SPI_TRANSCTRL_SLV_DATA_ONLY   (0x1 << 31)
#define APB_SPI_TRANSCTRL_CMDEN           (0x1 << 30)
#define APB_SPI_TRANSCTRL_ADDREN          (0x1 << 29)
#define APB_SPI_TRANSCTRL_ADDRFMT         (0x1 << 28)

// RD/WR transfer count
#define RD_TRANCNT_MSK                	  (0x1ff << 0)
#define RD_TRANCNT(num)                   ((num - 1) << 0)
#define WR_TRANCNT_MSK                    (0x1ff << 12)
#define WR_TRANCNT(num)                   ((num - 1) << 12)

// APB SPI transfer mode
#define APB_SPI_TRANSMODE_MSK             (0xf << 24)
#define APB_SPI_TRANSMODE_WRnRD           (0x0 << 24)
#define APB_SPI_TRANSMODE_WRONLY          (0x1 << 24)
#define APB_SPI_TRANSMODE_RDONLY          (0x2 << 24)
#define APB_SPI_TRANSMODE_WR_RD           (0x3 << 24)
#define APB_SPI_TRANSMODE_RD_WR           (0x4 << 24)
#define APB_SPI_TRANSMODE_WR_DMY_RD       (0x5 << 24)
#define APB_SPI_TRANSMODE_RD_DMY_WR       (0x6 << 24)
#define APB_SPI_TRANSMODE_NONEDATA        (0x7 << 24)
#define APB_SPI_TRANSMODE_DMY_WR          (0x8 << 24)
#define APB_SPI_TRANSMODE_DMY_RD          (0x9 << 24)

/* APB SPI control register */
#define SPIRST                            (1UL << 0)
#define RXFIFORST                         (1UL << 1)
#define TXFIFORST                         (1UL << 2)
#define RXDMAEN                           (1UL << 3)
#define TXDMAEN                           (1UL << 4)
#define RXTHRES(num)                      (num << 8)
#define TXTHRES(num)                      (num << 16)

#define THRES_MASK_FIFO_16                (0x1f)
#define THRES_MASK_FIFO_128               (0xff)
#define RXTHRES_OFFSET                    (8)
#define TXTHRES_OFFSET                    (16)

/* APB SPI interrupt status register */
/* APB SPI interrupt enable register */
#define APB_SPI_RXFIFOOORINT              (1UL << 0)
#define APB_SPI_TXFIFOOURINT              (1UL << 1)
#define APB_SPI_RXFIFOINT                 (1UL << 2)
#define APB_SPI_TXFIFOINT                 (1UL << 3)
#define APB_SPI_ENDINT                    (1UL << 4)
#define APB_SPI_SLVCMD                    (1UL << 5)

/* APB SPI Slave Data Count Register */
#define APB_SPI_SLAVE_TXCNT_POS           (16)
#define APB_SPI_SLAVE_TXCNT_MASK          (0x3ff << APB_SPI_SLAVE_TXCNT_POS)
#define APB_SPI_SLAVE_RXCNT_POS           (0)
#define APB_SPI_SLAVE_RXCNT_MASK          (0x3ff << APB_SPI_SLAVE_RXCNT_POS)

// APB SPI flags
#define APB_SPI_FLAG_INITIALIZED          (1UL << 0)     // APB SPI initialized
#define APB_SPI_FLAG_POWERED              (1UL << 1)     // APB SPI powered on
#define APB_SPI_FLAG_CONFIGURED           (1UL << 2)     // APB SPI configured
#define APB_SPI_FLAG_DATA_LOST            (1UL << 3)     // APB SPI data lost occurred
#define APB_SPI_FLAG_MODE_FAULT           (1UL << 4)     // APB SPI mode fault occurred

// APB SPI transfer information (Run-time)
typedef struct _APB_SPI_TRANSFER_INFO
{
	unsigned char			transfer_op;   		// Transfer operation: send/receive/transfer
	unsigned char			*rx_buf;       		// Pointer to in data buffer
	unsigned char			*tx_buf;       		// Pointer to out data buffer
	unsigned char			*tx_buf_limit; 		// Pointer to the end of in data buffer
	unsigned int			rx_cnt;        		// Number of data received
	unsigned int			tx_cnt;        		// Number of data sent
	unsigned char			txfifo_refill; 		// The size of data APB SPI TX ISR refills one time.
	unsigned char			dma_tx_complete; 	// DMA driver have handled it's completion, or APB SPI driver needs to help to close DMA channel.
	unsigned char			dma_rx_complete; 	// DMA driver have handled it's completion, or APB SPI driver needs to help to close DMA channel.
} APB_SPI_TRANSFER_INFO;

// APB SPI status
typedef struct _APB_SPI_STATUS
{
	unsigned char			busy;         		// Transmitter/receiver busy flag
	unsigned char			data_lost;   		// Data lost: receiver overflow/transmit underflow (cleared on start of transfer operation)
	unsigned char			mode_fault;   		// Mode fault detected; optional (cleared on start of transfer operation)
} APB_SPI_STATUS;

// APB SPI information (Run-time)
typedef struct _APB_SPI_INFO
{
	APB_SPI_SignalEvent_t	cb_event;      		// Event callback
	APB_SPI_STATUS			status;        		// APB SPI status flags
	APB_SPI_TRANSFER_INFO	xfer;          		// APB SPI transfer information
	unsigned char			flags;         		// APB SPI driver flags
	unsigned int			mode;          		// APB SPI mode
	unsigned char			is_header;     		// TX header flag
	unsigned char			*record_rx_buf;		// Record the original RX buffer (Only used when there is TX header)
	unsigned char			txfifo_size;   		// APB SPI HW TXFIFO size
	unsigned char			data_bits;     		// the size of one unit APB SPI data (1 ~ 32, defaults: 8 bits)
	unsigned char			src_width;     		// APB SPI transfer width(align 8 bits)
	unsigned int            data_num;      		// Number of the pure transfer data(Does not include header)
	unsigned char			tx_header_len; 		// APB SPI header(Usually include CMD, ADDRESS and DUMMY)
} APB_SPI_INFO;

// APB SPI DMA
#if (APB_SPI_DMA_EN == 1)
typedef const struct _APB_SPI_DMA
{
	unsigned char				channel;     	// DMA channel
	unsigned char				reqsel;      	// DMA request selection
	DMA_SignalEvent_t			cb_event;   	// DMA event callback
} APB_SPI_DMA;
#endif

// General power states
typedef enum _APB_SPI_POWER_STATE
{
	APB_SPI_POWER_OFF,                      // Power off: no operation possible
	APB_SPI_POWER_LOW,                      // Low power mode: retain state, detect and signal wake-up events
	APB_SPI_POWER_FULL                      // Power on: full operation at maximum performance
} APB_SPI_POWER_STATE;

// APB SPI resources definitions
typedef struct
{
	APB_SPI_CAPABILITIES		capabilities; // Capabilities
	APB_SPI_RegDef				*reg;         // Pointer to APB SPI peripheral
	unsigned int				irq_num;      // APB SPI IRQ number
#if (APB_SPI_DMA_EN == 1)
	APB_SPI_DMA					*dma_tx;      // APB SPI TX DMA
	APB_SPI_DMA					*dma_rx;      // APB SPI RX DMA
#endif
	APB_SPI_INFO				*info;        // APB SPI run-time information
} const APB_SPI_RESOURCES;



// *** Functions *** //

// Get capabilities
APB_SPI_CAPABILITIES apb_spix_get_capabilities(APB_SPI_RESOURCES *spi);

// Initializes APB SPI
int apb_spix_initialize(APB_SPI_SignalEvent_t cb_event, APB_SPI_RESOURCES *spi);

// Uninitialized
int apb_spix_uninitialize(APB_SPI_RESOURCES *spi);

// Control
int apb_spix_power_control(APB_SPI_POWER_STATE state, APB_SPI_RESOURCES *spi);

// Send data
int apb_spix_send(const void *data, unsigned int num, APB_SPI_RESOURCES *spi);

// Receive data
int apb_spix_receive(void *data, unsigned int num, APB_SPI_RESOURCES *spi);

// Transfer data
int apb_spix_transfer(const void *data_out, void *data_in, unsigned int num, APB_SPI_RESOURCES *spi);

// Get data count
int apb_spix_get_data_count(APB_SPI_RESOURCES *spi);

// Control
int apb_spix_control(unsigned int control, unsigned int arg, APB_SPI_RESOURCES *spi);

// Get status
APB_SPI_STATUS apb_spix_get_status(APB_SPI_RESOURCES *spi);

// Interrupt handler
void apb_spix_irq_handler(APB_SPI_RESOURCES *spi);

#if (APB_SPI_DMA_EN == 1)
// DMA TX event
void apb_spix_dma_tx_event (unsigned int event, APB_SPI_RESOURCES * spi);

// DMA RX event
void apb_spix_dma_rx_event (unsigned int event, APB_SPI_RESOURCES * spi);
#endif


#ifdef __cplusplus
}
#endif


#endif	/* __APB_SPI_DRIVER_H__ */
