/*
 * ******************************************************************************************
 *
 * File		: spi.c
 * Author	: GOWIN
 * Function	: APB SPI demo.
 *
 * ******************************************************************************************
 */


// Includes ---------------------------------------------------------------------------------
#include "spi.h"
#include "apb_spi_dev.h"

#include "delay.h"
#include "uart.h"
#include <stdio.h>


// Definitions ------------------------------------------------------------------------------

// For user defined
// Select SPI master mode or slave mode
#define SPI_MASTER_MODE     1						// Enable/disable SPI master mode
#define SPI_SLAVE_MODE      !(SPI_MASTER_MODE)		// Enable/disable SPI slave mode
// Data transmission direction
#define MASTER2SLAVE        0						// Send data from master to slave
#define SLAVE2MASTER        !(MASTER2SLAVE)			// Send data from slave to master

// SPI commands
#define SPI_READ    0x0b
#define SPI_WRITE   0x51
#define SPI_DUMMY   0xff

#define MTIME					PLMT_BASE
#define FAKE_CPUFREQ_S			PCLKFREQ
#define SLAVE_LATENCY_MS		0.5
#define DELAY_CYCLE				(SLAVE_LATENCY_MS * (FAKE_CPUFREQ_S/1000))

#define TEST_DATA_SIZE      	8
#define TOTAL_TRANSFER_SIZE 	(TEST_DATA_SIZE + 2) // Total transfer size is cmd(1) + dummy(1) + data(TEST_DATA_SIZE)

static volatile char spi_event_transfer_complete = 0;


static void wait_complete(void)
{
	while (!spi_event_transfer_complete);
	printf("%d \r\n", spi_event_transfer_complete);
	spi_event_transfer_complete = 0;
}

// SPI callback
void spi_callback(unsigned int event)
{
	switch (event)
	{
		case APB_SPI_EVENT_TRANSFER_COMPLETE:
			spi_event_transfer_complete = 1;
			break;
		case APB_SPI_EVENT_DATA_LOST:
			while(1);
			break;
		default:
			while(1);
			break;
	}
}

#if SPI_MASTER_MODE
// Test SPI master mode

#if MASTER2SLAVE
static void spi_prepare_data_out(unsigned char *data_out, unsigned char cmd, unsigned int dummy_cnt, unsigned char *w_data, unsigned int wcnt)
{
	unsigned char *ptr = data_out;
	unsigned int i;

	// Set 1-byte command
	*ptr++ = cmd;

	// Set n-byte dummy
	for (i = 0; i < dummy_cnt; i++)
	{
		*ptr++ = SPI_DUMMY;
	}

	// Set n-byte w_data
	for (i = 0; i < wcnt; i++)
	{
		*ptr++ = *w_data++;
	}
}
#endif	/* MASTER2SLAVE */

int spi_master_mode (void)
{
	unsigned char data_out[TOTAL_TRANSFER_SIZE] = {0};
	int i;

	// Initializes UART2
	uart_init(38400);	// Baud rate is 38400

	printf("\r\nIt's an APB SPI Master Mode demo.\r\n");
	printf("\r\n====Please hardwire 2 test board=====\r\n\r\n");

	// Initializes APB SPI1
	apb_spi1_initialize(spi_callback);

	// Power full the APB SPI1 peripheral
	apb_spi1_power_control(APB_SPI_POWER_FULL);

	// Configure the APB SPI1 to master, 8-bit data length and bus speed to 1MHz
	apb_spi1_control(APB_SPI_MODE_MASTER |
			 	  	 APB_SPI_CPOL0_CPHA0 |
					 APB_SPI_MSB_LSB |
					 APB_SPI_DATA_BITS(8), 1000000);
	apb_spi1_control(APB_SPI_TX_HEADER_LENGTH , 2);

	printf("Master write/read test...\r\n");

#if MASTER2SLAVE
	unsigned char w_data[TEST_DATA_SIZE] = {0};

	// Prepare write data
	for (i = 0; i < TEST_DATA_SIZE; i++)
	{
		w_data[i] = i;
	}

	spi_prepare_data_out(data_out, SPI_WRITE, 1, w_data, TEST_DATA_SIZE);

	// Write data to slave
	apb_spi1_send(data_out, TOTAL_TRANSFER_SIZE);
	wait_complete();

	for(int i = 0; i < TOTAL_TRANSFER_SIZE; i++)
	{
		printf("data_out[%d]=%d \r\n",i, data_out[i]);
	}
#elif SLAVE2MASTER
	unsigned char data_in[TEST_DATA_SIZE] = {0};

    for (i = 0; i < TOTAL_TRANSFER_SIZE; i++)
    {
    	data_out[i] = 0;
    }

	simple_delay(200);

	apb_spi1_transfer(data_out, data_in, TOTAL_TRANSFER_SIZE);
	wait_complete();

	for (i = 0; i < TEST_DATA_SIZE; i++)
	{
		printf("data_in[%d]=%d \r\n", i, data_in[i]);
	}
#endif

	return 0;
}

#elif SPI_SLAVE_MODE
// Test SPI slave mode

int spi_slave_mode (void)
{
	unsigned char data[TEST_DATA_SIZE] = {0};

	// Initializes UART2
    uart_init(38400);	// Baud rate is 38400

	printf("\r\nIt's an APB SPI Slave Mode demo.\r\n");
	printf("\r\n====Please hardwire 2 test board=====\r\n\r\n");

	// Initializes APB SPI1
	apb_spi1_initialize(spi_callback);

	// Power full the APB SPI1 peripheral
	apb_spi1_power_control(APB_SPI_POWER_FULL);

	// Configure the APB SPI1 to slave, 8-bit data length
	apb_spi1_control(APB_SPI_MODE_SLAVE |
			 	 	 APB_SPI_CPOL0_CPHA0 |
					 APB_SPI_MSB_LSB |
					 APB_SPI_DATA_BITS(8), 0);

	printf("Slave write/read test...\r\n");

#if MASTER2SLAVE
	//Read data from master
	apb_spi1_receive(data, TEST_DATA_SIZE);
	wait_complete();

	for(int i = 0; i < TEST_DATA_SIZE; i++)
	{
		printf("data[%d]=%d \r\n", i, data[i]);
	}
#elif SLAVE2MASTER
	for (int i = 0; i < TEST_DATA_SIZE; i++)
	{
		data[i] = i+1;
	}

	// Write data to master
	apb_spi1_send(data, TEST_DATA_SIZE);
	wait_complete();
	for(int i = 0; i < TEST_DATA_SIZE; i++)
	{
		printf("data[%d]=%d \r\n", i, data[i]);
	}
#endif

	return 0;
}

#endif

void apb_spi_demo(void)
{
#if SPI_MASTER_MODE
	// Master mode
	spi_master_mode();
#elif SPI_SLAVE_MODE
	// Slave mode
	spi_slave_mode();
#endif
}
